import datetime
import itertools
import json
import logging
import os
import sqlite3
from sqlite3 import DatabaseError
from typing import Optional, List, Dict, Tuple, Union, Type

import networkx as nx
import numpy as np
import overpy
import pandas as pd
from ipyleaflet import Map, ScaleControl, FullScreenControl, Polyline, Marker, Circle, LayerGroup
from ipywidgets import HTML
from pandas.io.sql import DatabaseError as PandasDatabaseError
from scipy.spatial import KDTree
from scipy.stats import norm

from pyridy import config
from pyridy.osm import OSM, OSMRailwayLine
from pyridy.osm.utils import is_point_within_line_projection, project_point_onto_line
from pyridy.utils import Sensor, AccelerationSeries, LinearAccelerationSeries, MagnetometerSeries, OrientationSeries, \
    GyroSeries, RotationSeries, GPSSeries, PressureSeries, HumiditySeries, TemperatureSeries, WzSeries, LightSeries, \
    SubjectiveComfortSeries, AccelerationUncalibratedSeries, MagnetometerUncalibratedSeries, GyroUncalibratedSeries, \
    GNSSClockMeasurementSeries, GNSSMeasurementSeries, NMEAMessageSeries, TimeSeries, NTPDatetimeSeries
from pyridy.utils.device import Device
from pyridy.utils.tools import generate_random_color

logger = logging.getLogger(__name__)


class RDYFile:
    def __init__(self, path: str = "", sync_method: str = "timestamp", cutoff: bool = True,
                 timedelta_unit: str = 'timedelta64[ns]',
                 strip_timezone: bool = True,
                 filename="",
                 series: Union[List[Type[TimeSeries]], Type[TimeSeries]] = None,
                 color: str = None):
        """

        Parameters
        ----------
        path: str
            Path to the Ridy File
        sync_method: str
            Sync method to be applied
        cutoff: bool, default: True
            If True, cutoffs the measurements precisely to the timestamp when the measurement was started, respectively
            stopped. By default Ridy measurement files can contain several seconds of measurements from before/after
            the button press
        timedelta_unit: str
            NumPy timedelta unit to applied
        strip_timezone: bool, default: True
            Strips timezone from timestamps as np.datetime64 does not support timezones
        filename: str
            Name of the files, will be the filename if not provided
        """
        self.path = path

        # Sanity check if series is arg is valid
        if series:
            if type(series) is list:
                for s in series:
                    if not issubclass(s, TimeSeries):
                        raise ValueError("%s in %s is not a TimeSeries!" % (type(s), list(series)))
                self._series = series
            elif issubclass(series, TimeSeries):
                self._series = [series]
                pass
            else:
                raise ValueError("series argument must be list of TimeSeries or TimeSeries! not %s" % type(series))
        else:
            self._series = None

        if sync_method is not None and sync_method not in ["timestamp", "seconds", "device_time", "gps_time",
                                                           "ntp_time"]:
            raise ValueError(
                "synchronize argument must 'timestamp', 'seconds','device_time', 'gps_time' or 'ntp_time' not %s" %
                sync_method)

        self.sync_method = sync_method
        self.cutoff = cutoff
        self.timedelta_unit = timedelta_unit
        self.strip_timezone = strip_timezone

        # Ridy App Info
        self.ridy_version: Optional[str] = None
        self.ridy_version_code: Optional[int] = None

        # Geographic extent of file (can only be determined if GPSSeries was recorded
        self.lat_sw, self.lon_sw = None, None
        self.lat_ne, self.lon_ne = None, None
        self.bbox = None

        # RDY File Infos
        self.rdy_format_version: Optional[float] = None
        self.rdy_info_name: Optional[str] = None
        self.rdy_info_sex: Optional[str] = None
        self.rdy_info_age: Optional[int] = None
        self.rdy_info_height: Optional[float] = None
        self.rdy_info_weight: Optional[float] = None

        self.t0: Optional[np.datetime64] = None
        self.cs_matrix_string: Optional[str] = None
        self.cs_matrix: Optional[np.ndarray] = None  # TODO
        self.timestamp_when_started: Optional[int] = None
        self.timestamp_when_stopped: Optional[int] = None
        self.ntp_timestamp: Optional[int] = None
        self.ntp_date_time: Optional[np.datetime64] = None
        self.device: Optional[Device] = None

        self.duration: Optional[float] = None

        # Sensors
        self.sensors: Optional[List[Sensor]] = []

        # Measurement Series
        self.measurements = {AccelerationSeries: AccelerationSeries(),
                             AccelerationUncalibratedSeries: AccelerationUncalibratedSeries(),
                             LinearAccelerationSeries: LinearAccelerationSeries(),
                             MagnetometerSeries: MagnetometerSeries(),
                             MagnetometerUncalibratedSeries: MagnetometerUncalibratedSeries(),
                             OrientationSeries: OrientationSeries(),
                             GyroSeries: GyroSeries(),
                             GyroUncalibratedSeries: GyroUncalibratedSeries(),
                             RotationSeries: RotationSeries(),
                             GPSSeries: GPSSeries(),
                             GNSSClockMeasurementSeries: GNSSClockMeasurementSeries(),
                             GNSSMeasurementSeries: GNSSMeasurementSeries(),
                             NMEAMessageSeries: NMEAMessageSeries(),
                             PressureSeries: PressureSeries(),
                             TemperatureSeries: TemperatureSeries(),
                             HumiditySeries: HumiditySeries(),
                             LightSeries: LightSeries(),
                             WzSeries: WzSeries(),
                             SubjectiveComfortSeries: SubjectiveComfortSeries(),
                             NTPDatetimeSeries: NTPDatetimeSeries()}

        # Filename and extension
        self.filename: Optional[str] = filename
        self.extension: Optional[str] = ""

        # OSM Data (set by Campaign)
        self.osm: Optional[OSM] = None
        self.matched_nodes: Optional[List[overpy.Node]] = []  # Nodes from Map Matching
        self.matched_ways: Optional[List[overpy.Way]] = []  # Ways from Map Matching
        self.matched_line: Optional[OSMRailwayLine] = None  # Matched Railway Line

        if self.path:
            self.load_file(self.path)

            if self.timestamp_when_started and self.timestamp_when_stopped:
                self.duration = (self.timestamp_when_stopped - self.timestamp_when_started) * 1e-9

            if self.sync_method:
                self._synchronize()

            if len(self.measurements[GPSSeries]) > 0:
                self.lon_sw = self.measurements[GPSSeries].lon.min()
                self.lon_ne = self.measurements[GPSSeries].lon.max()

                self.lat_sw = self.measurements[GPSSeries].lat.min()
                self.lat_ne = self.measurements[GPSSeries].lat.max()

                self.bbox = [self.lon_sw, self.lat_sw, self.lon_ne, self.lat_ne]
        else:
            logging.warning("RDYFile instantiated without a path")

        # Unique color for each line
        if not color:
            while True:
                self.color = generate_random_color("HEX")
                if self.color not in config.colors:
                    config.colors.append(color)
                    break
                else:
                    continue
        else:
            self.color = color

    # def __getitem__(self, idx):
    #     key = list(self.measurements.keys())[idx]
    #     return self.measurements[key]

    def __iter__(self):
        """

        Returns
        -------
            FileIterator
        """
        return FileIterator(self)

    def __repr__(self):
        return "Filename: %s, T0: %s, Duration: %s" % (self.filename,
                                                       str(self.t0),
                                                       str(datetime.timedelta(seconds=self.duration)))

    def _do_candidate_search(self, osm_xy: np.ndarray, track_xy: np.ndarray, hor_acc: np.ndarray):
        """ Internal method to search for candidate edges for map matching

        """
        # Find the closest coordinates using KDTrees
        kd_tree_osm = KDTree(osm_xy)
        kd_tree_track = KDTree(track_xy)

        # Indices of OSM nodes that are close to each respective GPS point within radius r
        indices = kd_tree_track.query_ball_tree(kd_tree_osm, r=100)

        c_dict = {}  # Dict with node candidates for each GPS coord
        edges = []  # Candidate edges with emission probabilities

        # Perform search for node candidate on all GPS coords
        for i, idxs in enumerate(indices):
            # Get unique ways based on indices
            c_ways = list(set(list(itertools.chain(*[self.osm.nodes[idx].ways for idx in idxs]))))

            # Find candidate line segments
            c_segs = []
            for w in c_ways:
                n = w.nodes
                segs = []  # List of suitable line segments

                for n1, n2 in zip(n, n[1:]):
                    x1, y1 = n1.attributes["x"], n1.attributes["y"]
                    x2, y2 = n2.attributes["x"], n2.attributes["y"]

                    # Only take those line segment into consideration where the perpendicular projection
                    # of the GPS coords lies inside the line segment
                    b = is_point_within_line_projection(line=[[x1, y1], [x2, y2]], point=track_xy[i])
                    if b:
                        # Point of orthogonal intersection
                        p, d = project_point_onto_line(line=[[x1, y1], [x2, y2]], point=track_xy[i])
                        if d < hor_acc[i]:
                            p_lon, p_lat = self.osm.utm_proj(p[0], p[1], inverse=True)
                            segs.append([d, p_lon, p_lat, n1, n2, w.id, None, i])

                if segs:
                    segs = np.array(segs)
                    i_min = np.argmin(
                        segs[:, 0])  # Select candidate line segment based on smallest perpendicular distance
                    n1, n2 = segs[i_min, 3], segs[i_min, 4]

                    e1 = list(self.osm.G.edges(n1.id, keys=True))
                    e2 = list(self.osm.G.edges(n2.id, keys=True))

                    inter = list(set(e1).intersection(e2))
                    if len(inter) == 0:  # TODO
                        c_seg_e = e1[0]
                    else:
                        c_seg_e = inter[0]

                    c_seg = segs[i_min]
                    c_segs.append(c_seg)

                    edges.append([c_seg_e, hor_acc[i], c_seg[0]])

            c_dict[i] = {"c_ways": c_ways, "c_segs": c_segs}

        # Calculate emission probabilities for each edge candidate
        c_edges = {}
        if edges:
            edges = np.array(edges, dtype='object')
            e_probs = norm.pdf(edges[:, 2].astype(float), np.zeros(len(edges)), edges[:, 1].astype(float))

            for i, e_prob in enumerate(e_probs):
                if edges[i][0] not in c_edges:
                    c_edges[edges[i][0]] = {"e_prob": [e_prob]}
                else:
                    c_edges[edges[i][0]]["e_prob"].append(e_prob)

        return c_dict, c_edges

    def _synchronize(self):
        """ Internal method that synchronizes the timestamps to a given sync timestamp

        """
        if self.sync_method == "timestamp":
            for m in self.measurements.values():
                m.synchronize("timestamp", self.timestamp_when_started,
                              timedelta_unit=self.timedelta_unit)
        elif self.sync_method == 'seconds':
            for m in self.measurements.values():
                m.synchronize("seconds", self.timestamp_when_started,
                              timedelta_unit=self.timedelta_unit)
        elif self.sync_method == "device_time":
            if self.t0:
                for m in self.measurements.values():
                    m.synchronize("device_time", self.timestamp_when_started, self.t0,
                                  timedelta_unit=self.timedelta_unit)
            else:
                logger.warning("(%s) t0 is None, falling back to timestamp synchronization" % self.filename)
                self.sync_method = "timestamp"
                self._synchronize()
        elif self.sync_method == "gps_time":
            if len(self.measurements[GPSSeries]) > 0:
                sync_timestamp = self.measurements[GPSSeries].time[0]
                utc_sync_time = self.measurements[GPSSeries].utc_time[0]

                for i, t in enumerate(self.measurements[
                                          GPSSeries].utc_time):
                    # The first utc_time value ending with 000 is a real GPS measurement
                    if str(t)[-3:] == "000":
                        utc_sync_time = t
                        sync_timestamp = self.measurements[GPSSeries].time[i]
                        break

                sync_time = np.datetime64(int(utc_sync_time * 1e6), "ns")
                for m in self.measurements.values():
                    m.synchronize("gps_time", sync_timestamp, sync_time, timedelta_unit=self.timedelta_unit)
            else:
                logger.warning(
                    "(%s) No GPS time recording, falling back to device_time synchronization" % self.filename)
                self.sync_method = "device_time"
                self._synchronize()
        elif self.sync_method == "ntp_time":
            if self.ntp_timestamp and self.ntp_date_time:
                for m in self.measurements.values():
                    m.synchronize("ntp_time", self.ntp_timestamp, self.ntp_date_time,
                                  timedelta_unit=self.timedelta_unit)
            else:
                logger.warning("(%s) No ntp timestamp and datetime, falling back to device_time synchronization" %
                               self.filename)

                self.sync_method = "device_time"
                self._synchronize()
        else:
            raise ValueError(
                "sync_method must 'timestamp', 'device_time', 'gps_time' or 'ntp_time' not %s" % self.sync_method)
        pass

    def create_map(self, t_lim: Tuple[np.datetime64, np.datetime64] = None, show_hor_acc: bool = False) -> Map:
        """ Creates an ipyleaflet Map using OpenStreetMap and OpenRailwayMap to show the GPS track of the
        measurement file

        Parameters
        ----------
            t_lim: tuple, default: None
                Time limit as a tuple of np.datetime64 to show only parts of the GPS track that are within the specified
                time interval
            show_hor_acc : bool, default: False
                If true shows the horizontal accuracies for each measurement point using circles. The likelihood that
                that the real position is within the circle is defined as 68 %

        Returns
        -------
            Map
        """
        gps_series = self.measurements[GPSSeries]
        coords = gps_series.to_ipyleaflef()
        time = gps_series.time
        hor_acc = gps_series.hor_acc

        if coords == [[]]:
            logger.warning("(%s) Cant create map, GPSSeries is empty!" % self.filename)
        else:
            if t_lim:
                if type(t_lim) != tuple:
                    raise ValueError("t_lim must be a tuple of np.datetime64")

                if t_lim[0] > t_lim[1]:
                    raise ValueError("The first datetime for t_lim must be smaller than the second!")

                mask = (gps_series.time >= t_lim[0]) & (gps_series.time <= t_lim[1])

                coords = [c for i, c in enumerate(coords) if mask[i]]
                time = [t for i, t in enumerate(gps_series.time) if mask[i]]
                hor_acc = [h for i, h in enumerate(gps_series.hor_acc) if mask[i]]

            color = generate_random_color("HEX")

            m = Map(center=self.determine_track_center()[::-1],
                    zoom=12,
                    scroll_wheel_zoom=True,
                    basemap=config.OPEN_STREET_MAP_DE)

            m.add_control(ScaleControl(position='bottomleft'))
            m.add_control(FullScreenControl())

            # Add map
            m.add_layer(config.OPEN_RAILWAY_MAP)

            file_polyline = Polyline(locations=coords, color=color, fill=False, weight=4, dash_array='10, 10')
            m.add_layer(file_polyline)

            start_marker = Marker(location=tuple(coords[0]), draggable=False, icon=config.START_ICON)
            end_marker = Marker(location=tuple(coords[-1]), draggable=False, icon=config.END_ICON)

            start_message = HTML()
            end_message = HTML()
            start_message.value = "<p>Start:</p><p>" + self.filename + "</p><p>" \
                                  + str(time[0] or 'n/a') + "</p><p>" \
                                  + str(getattr(self.device, "manufacturer", "n/a")) + "; " \
                                  + str(getattr(self.device, "model", "n/a")) + "</p>"

            end_message.value = "<p>End:</p><p>" + self.filename + "</p><p>" \
                                + str(time[-1] or 'n/a') + "</p><p>" \
                                + str(getattr(self.device, "manufacturer", "n/a")) + "; " \
                                + str(getattr(self.device, "model", "n/a")) + "</p>"

            start_marker.popup = start_message
            end_marker.popup = end_message

            m.add_layer(start_marker)
            m.add_layer(end_marker)

            if show_hor_acc:
                circles = []
                for c, h in zip(coords, hor_acc):
                    circle = Circle()
                    circle.location = (c[0], c[1])
                    circle.radius = int(h)
                    circle.color = "#00549F"
                    circle.fill_color = "#00549F"
                    circle.weight = 3
                    circle.fill_opacity = 0.1

                    circles.append(circle)

                l_circles = LayerGroup(layers=circles)
                m.add_layer(l_circles)

            return m

    def determine_track_center(self, gps_series: Optional[GPSSeries] = None) -> (float, float):
        """ Determines the geographical center of the GPSSeries, returns None if the GPSSeries is emtpy.

        Parameters
        ----------
            gps_series: GPSSeries, default: None
                If not None, takes the given GPSSeries to determine the track center

        Returns
        -------
            float, float
        """
        if not gps_series:
            gps_series = self.measurements[GPSSeries]

        if gps_series.is_empty():
            logger.warning("(%s) Cant determine track center, GPSSeries is empty!" % self.filename)
        else:
            center_lon = (gps_series.lon.max() + gps_series.lon.min()) / 2
            center_lat = (gps_series.lat.max() + gps_series.lat.min()) / 2

            logging.info("Geographic center of track: Lon: %s, Lat: %s" % (str(center_lon), str(center_lat)))

            return center_lon, center_lat

    def do_map_matching(self, v_thres: float = config.options["MAP_MATCHING_V_THRES"],
                        algorithm: str = config.options["MAP_MATCHING_DEFAULT_ALGORITHM"],
                        alpha: int = config.options["MAP_MATCHING_ALPHA"],
                        beta: int = config.options["MAP_MATCHING_BETA"]):
        """ Performs map matching of the GPS track to closest OSM nodes/ways

        Parameters
        ----------
        alpha: float, default: 1.0
            Weighting Parameter for the Map Matching Algorithm. Alpha represents the default lange of an edge in the
            Graph
        beta: float, default: 1.0
            Beta is a scaling factor for the emission probabilities.
        algorithm: str, default: nx
            Algorithm to be used, can be "pyridy" or "nx". The pyridy algorithm also incorporates how switches
            can be transited
        v_thres: float, default: 1.0
            Speed threshold, GPS points measured with a velocity below v_thres [m/s] will not be considered

        Returns
        -------

        """
        if algorithm not in ["nx", "pyridy"]:
            raise ValueError("(%s) Algorithm must be either nx or pyridy, not %s" % (self.filename, algorithm))

        if self.osm:
            # Prepare data
            gps_coords = self.measurements[GPSSeries]
            lon = gps_coords.lon[gps_coords.speed > v_thres]
            lat = gps_coords.lat[gps_coords.speed > v_thres]
            hor_acc = gps_coords.hor_acc[gps_coords.speed > v_thres]

            n_gps = len(lon)

            x, y = self.osm.utm_proj(lon, lat)

            track_xy = np.vstack([x, y]).T
            osm_xy = self.osm.get_coords(frmt="xy")

            c_dict, c_edges = self._do_candidate_search(osm_xy, track_xy, hor_acc)

            # Initialize edge weights
            for e in self.osm.G.edges:
                self.osm.G.edges[e]["c_weight"] = 1 * alpha

            for k in c_edges.keys():
                self.osm.G.edges[k]["c_weight"] = 1 / (1 + beta * sum(c_edges[k]["e_prob"]))

            # Perform map matching
            s_n = None
            for i in range(n_gps):
                if len(c_dict[i]["c_segs"]) > 0:
                    s_n = c_dict[i]["c_segs"][np.array(c_dict[i]["c_segs"])[:, 0].argmin()][3].id
                    break

            e_n = None
            for i in reversed(range(n_gps)):
                if len(c_dict[i]["c_segs"]) > 0:
                    e_n = c_dict[i]["c_segs"][np.array(c_dict[i]["c_segs"])[:, 0].argmin()][4].id
                    break

            if not s_n or not e_n:
                logger.warning("(%s) Map matching failed, no start or end node found!" % self.filename)
            else:
                # Use Dijkstra's shortest path to perform map matching, but use a weighting based on emission
                # probabilities instead of node distances
                if algorithm == 'pyridy':
                    # Matched node ids
                    _, _, m_n_ids = self.osm.get_shortest_path(source=s_n, target=e_n, weight="c_weight")
                else:
                    m_n_ids = nx.shortest_path(self.osm.G, source=s_n, target=e_n, weight="c_weight")

                self.matched_nodes = [self.osm.node_dict[n] for n in m_n_ids]  # Matched nodes

                m_w_ids = [self.osm.G[n1][n2][0]["way_id"] for n1, n2 in zip(m_n_ids, m_n_ids[1:])]
                self.matched_ways = list(set([self.osm.way_dict[w_id] for w_id in m_w_ids]))  # Mapped Ways
                logger.debug("(%s) Found %d nodes that match the files GPS track!" % (self.filename,
                                                                                      len(self.matched_nodes)))

                if len(self.matched_nodes) > 0:
                    m_ratios = [len(set(self.matched_nodes).intersection(line.nodes)) / len(self.matched_nodes) for line
                                in
                                self.osm.railway_lines if self]

                    if max(m_ratios) > config.options["MAP_MATCHING_MIN_LINE_MATCH_RATIO"]:
                        self.matched_line = self.osm.railway_lines[m_ratios.index(max(m_ratios))]
                        logger.debug("(%s) Matched Railway Line: %s (%.f Match Ratio)" % (self.filename,
                                                                                          self.matched_line.name,
                                                                                          max(m_ratios)))
                    else:
                        logger.debug("(%s) Could not match map matching results to a specific railway line" %
                                     self.filename)
                else:
                    logger.debug("(%s) Could not match map matching results to a specific railway line" %
                                 self.filename)
        else:
            logger.warning("(%s) Can't do map matching since no file contains no OSM data" % self.filename)
        pass

    def get_integrity_report(self):
        """ Returns a dict that contains information which measurement types are available in the file


        Returns
        -------
            dict
        """
        report = {}

        for k, v in self.measurements.items():
            if len(v) > 0:
                report[k.__name__] = True
            else:
                report[k.__name__] = False

        attr = self.__dict__.copy()
        attr.update(attr["device"].__dict__.copy())
        for a in ["db_con", "measurements", "device", "sensors"]:
            attr.pop(a)
            pass

        report.update(attr)

        return report

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, v):
        for m in self.measurements.values():
            m.filename = v

        self._filename = v

    def load_file(self, path: str):
        """ Loads a single Ridy file located at path

        Parameters
        ----------
        path Path to the ridy file

        """
        logger.debug("Loading file: %s" % path)

        _, self.extension = os.path.splitext(path)
        _, self.filename = os.path.split(path)

        if self.extension == ".rdy":
            with open(path, 'r') as file:
                rdy = json.load(file)

            if 'Ridy_Version' in rdy:
                self.ridy_version = rdy['Ridy_Version']
            else:
                logger.debug("No Ridy_Version in file: %s" % self.filename)
                self.ridy_version = None

            if 'Ridy_Version_Code' in rdy:
                self.ridy_version_code = rdy['Ridy_Version_Code']
            else:
                logger.debug("No Ridy_Version_Code in file: %s" % self.filename)
                self.ridy_version_code = None

            if 'RDY_Format_Version' in rdy:
                self.rdy_format_version = rdy['RDY_Format_Version']
            else:
                logger.debug("No RDY_Format_Version in file: %s" % self.filename)
                self.rdy_format_version = None

            if 'RDY_Info_Name' in rdy:
                self.rdy_info_name = rdy['RDY_Info_Name']
            else:
                logger.debug("No RDY_Info_Name in file: %s" % self.filename)
                self.rdy_info_name = None

            if 'RDY_Info_Sex' in rdy:
                self.rdy_info_sex = rdy['RDY_Info_Sex']
            else:
                logger.debug("No RDY_Info_Sex in file: %s" % self.filename)
                self.rdy_info_sex = None

            if 'RDY_Info_Age' in rdy:
                self.rdy_info_age = rdy['RDY_Info_Age']
            else:
                logger.debug("No RDY_Info_Age in file: %s" % self.filename)
                self.rdy_info_age = None

            if 'RDY_Info_Height' in rdy:
                self.rdy_info_height = rdy['RDY_Info_Height']
            else:
                logger.debug("No RDY_Info_Height in file: %s" % self.filename)
                self.rdy_info_height = None

            if 'RDY_Info_Weight' in rdy:
                self.rdy_info_weight = rdy['RDY_Info_Weight']
            else:
                logger.debug("No RDY_Info_Weight in file: %s" % self.filename)
                self.rdy_info_weight = None

            if 't0' in rdy:
                if self.strip_timezone:
                    t0 = datetime.datetime.fromisoformat(rdy['t0']).replace(tzinfo=None)
                    self.t0 = np.datetime64(t0)
                else:
                    self.t0 = np.datetime64(rdy['t0'])
            else:
                self.t0 = None
                logger.debug("No t0 in file: %s" % self.filename)

            if 'cs_matrix_string' in rdy:
                self.cs_matrix_string = rdy['cs_matrix_string']
            else:
                self.cs_matrix_string = None
                logger.debug("No t0 in file: %s" % self.filename)

            if 'timestamp_when_started' in rdy:
                self.timestamp_when_started = rdy['timestamp_when_started']
            else:
                self.timestamp_when_started = None
                logger.debug("No timestamp_when_started in file: %s" % self.filename)

            if 'timestamp_when_stopped' in rdy:
                self.timestamp_when_stopped = rdy['timestamp_when_stopped']
            else:
                self.timestamp_when_stopped = None
                logger.debug("No timestamp_when_stopped in file: %s" % self.filename)

            if 'ntp_timestamp' in rdy:
                self.ntp_timestamp = rdy['ntp_timestamp']
            else:
                self.ntp_timestamp = None
                logger.debug("No ntp_timestamp in file: %s" % self.filename)

            if 'ntp_date_time' in rdy:
                if self.strip_timezone:
                    ntp_datetime_str = rdy['ntp_date_time']
                    if ntp_datetime_str:
                        ntp_date_time = datetime.datetime.fromisoformat(ntp_datetime_str).replace(tzinfo=None)
                        self.ntp_date_time = np.datetime64(ntp_date_time)
                    else:
                        self.ntp_date_time = None
                else:
                    self.ntp_date_time = np.datetime64(rdy['t0'])
            else:
                self.ntp_date_time = None
                logger.debug("No ntp_date_time in file: %s" % self.filename)

            if "device" in rdy:
                self.device = Device(**rdy['device_info'])
            else:
                self.device = Device()
                logger.debug("No device information in file: %s" % self.filename)

            if "sensors" in rdy:
                for sensor in rdy['sensors']:
                    self.sensors.append(Sensor(**sensor))
            else:
                logger.debug("No sensor descriptions in file: %s" % self.filename)

            if (self._series is not None and AccelerationSeries in self._series) or self._series is None:
                if "acc_series" in rdy:
                    self.measurements[AccelerationSeries] = \
                        AccelerationSeries(filename=self.filename,
                                           rdy_format_version=self.rdy_format_version,
                                           **rdy['acc_series'])
                else:
                    logger.debug("No Acceleration Series in file: %s" % self.filename)

            if (self._series is not None and AccelerationUncalibratedSeries in self._series) or self._series is None:
                if "acc_uncal_series" in rdy:
                    self.measurements[AccelerationUncalibratedSeries] = AccelerationUncalibratedSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **rdy['acc_uncal_series'])
                else:
                    logger.debug("No uncalibrated Acceleration Series in file: %s" % self.filename)

            if (self._series is not None and LinearAccelerationSeries in self._series) or self._series is None:
                if "lin_acc_series" in rdy:
                    self.measurements[LinearAccelerationSeries] = LinearAccelerationSeries(
                        filename=self.filename,
                        rdy_format_version=self.rdy_format_version,
                        **rdy['lin_acc_series'])
                else:
                    logger.debug("No Linear Acceleration Series in file: %s" % self.filename)

            if (self._series is not None and MagnetometerSeries in self._series) or self._series is None:
                if "mag_series" in rdy:
                    self.measurements[MagnetometerSeries] = \
                        MagnetometerSeries(filename=self.filename,
                                           rdy_format_version=self.rdy_format_version,
                                           **rdy['mag_series'])
                else:
                    logger.debug("No Magnetometer Series in file: %s" % self.filename)

            if (self._series is not None and MagnetometerUncalibratedSeries in self._series) or self._series is None:
                if "mag_uncal_series" in rdy:
                    self.measurements[MagnetometerUncalibratedSeries] = MagnetometerUncalibratedSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **rdy['mag_uncal_series'])
                else:
                    logger.debug("No uncalibrated Magnetometer Series in file: %s" % self.filename)

            if (self._series is not None and OrientationSeries in self._series) or self._series is None:
                if "orient_series" in rdy:
                    self.measurements[OrientationSeries] = OrientationSeries(filename=self.filename,
                                                                             rdy_format_version=self.rdy_format_version,
                                                                             **rdy['orient_series'])
                else:
                    logger.debug("No Orientation Series in file: %s" % self.filename)

            if (self._series is not None and GyroSeries in self._series) or self._series is None:
                if "gyro_series" in rdy:
                    self.measurements[GyroSeries] = GyroSeries(filename=self.filename,
                                                               rdy_format_version=self.rdy_format_version,
                                                               **rdy['gyro_series'])
                else:
                    logger.debug("No Gyro Series in file: %s" % self.filename)

            if (self._series is not None and GyroUncalibratedSeries in self._series) or self._series is None:
                if "gyro_uncal_series" in rdy:
                    self.measurements[GyroUncalibratedSeries] = GyroUncalibratedSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **rdy['gyro_uncal_series'])
                else:
                    logger.debug("No uncalibrated Gyro Series in file: %s" % self.filename)

            if (self._series is not None and RotationSeries in self._series) or self._series is None:
                if "rot_series" in rdy:
                    self.measurements[RotationSeries] = RotationSeries(filename=self.filename,
                                                                       rdy_format_version=self.rdy_format_version,
                                                                       **rdy['rot_series'])
                else:
                    logger.debug("No Rotation Series in file: %s" % self.filename)

            if (self._series is not None and GPSSeries in self._series) or self._series is None:
                if "gps_series" in rdy:
                    self.measurements[GPSSeries] = GPSSeries(filename=self.filename,
                                                             rdy_format_version=self.rdy_format_version,
                                                             **rdy['gps_series'])
                else:
                    logger.debug("No GPS Series in file: %s" % self.filename)

            if (self._series is not None and GNSSMeasurementSeries in self._series) or self._series is None:
                if "gnss_series" in rdy:
                    self.measurements[GNSSMeasurementSeries] = GNSSMeasurementSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **rdy['gnss_series'])
                else:
                    logger.debug("No GPS Series in file: %s" % self.filename)

            if (self._series is not None and GNSSClockMeasurementSeries in self._series) or self._series is None:
                if "gnss_clock_series" in rdy:
                    self.measurements[GNSSClockMeasurementSeries] = GNSSClockMeasurementSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **rdy['gnss_clock_series'])
                else:
                    logger.debug("No GNSS Clock Series in file: %s" % self.filename)

            if (self._series is not None and NMEAMessageSeries in self._series) or self._series is None:
                if "nmea_series" in rdy:
                    self.measurements[NMEAMessageSeries] = NMEAMessageSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **rdy['nmea_series'])
                else:
                    logger.debug("No GPS Series in file: %s" % self.filename)

            if (self._series is not None and PressureSeries in self._series) or self._series is None:
                if "pressure_series" in rdy:
                    self.measurements[PressureSeries] = PressureSeries(filename=self.filename,
                                                                       rdy_format_version=self.rdy_format_version,
                                                                       **rdy['pressure_series'])
                else:
                    logger.debug("No Pressure Series in file: %s" % self.filename)

            if (self._series is not None and TemperatureSeries in self._series) or self._series is None:
                if "temperature_series" in rdy:
                    self.measurements[TemperatureSeries] = TemperatureSeries(filename=self.filename,
                                                                             rdy_format_version=self.rdy_format_version,
                                                                             **rdy['temperature_series'])
                else:
                    logger.debug("No Temperature Series in file: %s" % self.filename)

            if (self._series is not None and HumiditySeries in self._series) or self._series is None:
                if "humidity_series" in rdy:
                    self.measurements[HumiditySeries] = HumiditySeries(filename=self.filename,
                                                                       rdy_format_version=self.rdy_format_version,
                                                                       **rdy['humidity_series'])
                else:
                    logger.debug("No Humidity Series in file: %s" % self.filename)

            if (self._series is not None and LightSeries in self._series) or self._series is None:
                if "light_series" in rdy:
                    self.measurements[LightSeries] = LightSeries(filename=self.filename,
                                                                 rdy_format_version=self.rdy_format_version,
                                                                 **rdy['light_series'])
                else:
                    logger.debug("No Light Series in file: %s" % self.filename)

            if (self._series is not None and WzSeries in self._series) or self._series is None:
                if "wz_series" in rdy:
                    self.measurements[WzSeries] = WzSeries(filename=self.filename,
                                                           rdy_format_version=self.rdy_format_version,
                                                           **rdy['wz_series'])
                else:
                    logger.debug("No Wz Series in file: %s" % self.filename)

            if (self._series is not None and SubjectiveComfortSeries in self._series) or self._series is None:
                if "subjective_comfort_series" in rdy:
                    self.measurements[SubjectiveComfortSeries] = SubjectiveComfortSeries(
                        filename=self.filename,
                        rdy_format_version=self.rdy_format_version,
                        **rdy['subjective_comfort_series'])
                else:
                    logger.debug("No Subjective Comfort Series in file: %s" % self.filename)

            if (self._series is not None and NTPDatetimeSeries in self._series) or self._series is None:
                if "ntp_series" in rdy:
                    self.measurements[NTPDatetimeSeries] = NTPDatetimeSeries(
                        filename=self.filename,
                        rdy_format_version=self.rdy_format_version,
                        strip_timezone=self.strip_timezone,
                        **rdy['ntp_series'])
                else:
                    logger.debug("No NTP Datetime Series in file: %s" % self.filename)

        elif self.extension == ".sqlite":
            db_con = sqlite3.connect(path)

            try:
                info: Dict = dict(pd.read_sql_query("SELECT * from measurement_information_table", db_con))
            except (DatabaseError, PandasDatabaseError) as e:
                logger.error(e)
                # Older files can contain wrong table name
                try:
                    info = dict(pd.read_sql_query("SELECT * from measurment_information_table", db_con))
                    logger.debug("(%s) Older file containing measu(rm)ent_information_table" % self.filename)
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing measurement_information_table" % self.filename)
                    logger.debug(e)
                    info = {}

            try:
                sensor_df = pd.read_sql_query("SELECT * from sensor_descriptions_table", db_con)
                for _, row in sensor_df.iterrows():
                    self.sensors.append(Sensor(**dict(row)))
            except (DatabaseError, PandasDatabaseError) as e:
                logger.debug(
                    "(%s) DatabaseError occurred when accessing sensor_descriptions_table" % self.filename)
                logger.debug(e)

            try:
                device_df = pd.read_sql_query("SELECT * from device_information_table", db_con)
                self.device = Device(**dict(device_df))
            except (DatabaseError, PandasDatabaseError) as e:
                logger.debug("(%s) DatabaseError occurred when accessing device_information_table" % self.filename)
                logger.debug(e)
                self.device = Device()

            # Info
            if 'ridy_version' in info and len(info['ridy_version']) > 1:
                logger.debug("Measurement information table contains more than 1 row!")

            if 'ridy_version' in info and len(info['ridy_version']) > 0:
                self.ridy_version = info['ridy_version'].iloc[-1]

            if 'ridy_version_code' in info and len(info['ridy_version_code']) > 0:
                self.ridy_version_code = info['ridy_version_code'].iloc[-1]

            if 'rdy_format_version' in info and len(info['rdy_format_version']) > 0:
                self.rdy_format_version = info['rdy_format_version'].iloc[-1]

            if 'rdy_info_name' in info and len(info['rdy_info_name']) > 0:
                self.rdy_info_name = info['rdy_info_name'].iloc[-1]

            if 'rdy_info_sex' in info and len(info['rdy_info_sex']) > 0:
                self.rdy_info_sex = info['rdy_info_sex'].iloc[-1]

            if 'rdy_info_age' in info and len(info['rdy_info_age']) > 0:
                self.rdy_info_age = info['rdy_info_age'].iloc[-1]

            if 'rdy_info_height' in info and len(info['rdy_info_height']) > 0:
                self.rdy_info_height = info['rdy_info_height'].iloc[-1]

            if 'rdy_info_weight' in info and len(info['rdy_info_weight']) > 0:
                self.rdy_info_weight = info['rdy_info_weight'].iloc[-1]

            if 't0' in info and len(info['t0']) > 0:
                if self.strip_timezone:
                    t0 = datetime.datetime.fromisoformat(info['t0'].iloc[-1]).replace(tzinfo=None)
                    self.t0 = np.datetime64(t0)
                else:
                    self.t0 = np.datetime64(info['t0'].iloc[-1])

            if 'cs_matrix_string' in info and len(info['cs_matrix_string']) > 0:
                self.cs_matrix_string = info['cs_matrix_string'].iloc[-1]

            if 'timestamp_when_started' and len(info['timestamp_when_started']) > 0:
                self.timestamp_when_started = info['timestamp_when_started'].iloc[-1]

            if 'timestamp_when_stopped' in info and len(info['timestamp_when_stopped']) > 0:
                self.timestamp_when_stopped = info['timestamp_when_stopped'].iloc[-1]

            if 'ntp_timestamp' in info and len(info['ntp_timestamp']) > 0:
                self.ntp_timestamp = info['ntp_timestamp'].iloc[-1]

            if 'ntp_date_time' in info and len(info['ntp_date_time']) > 0:
                if self.strip_timezone:
                    ntp_datetime_str = info['ntp_date_time'].iloc[-1]
                    if ntp_datetime_str:
                        ntp_date_time = datetime.datetime.fromisoformat(ntp_datetime_str).replace(tzinfo=None)
                        self.ntp_date_time = np.datetime64(ntp_date_time)
                    else:
                        self.ntp_date_time = None
                else:
                    self.ntp_date_time = np.datetime64(info['ntp_date_time'].iloc[-1])

            # Measurements
            if (self._series is not None and AccelerationSeries in self._series) or self._series is None:
                try:
                    acc_df = pd.read_sql_query("SELECT * from acc_measurements_table", db_con)
                    self.measurements[AccelerationSeries] = \
                        AccelerationSeries(filename=self.filename,
                                           rdy_format_version=self.rdy_format_version,
                                           **dict(acc_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing acc_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and AccelerationUncalibratedSeries in self._series) or self._series is None:
                try:
                    acc_uncal_df = pd.read_sql_query("SELECT * from acc_uncal_measurements_table", db_con)
                    self.measurements[AccelerationUncalibratedSeries] = AccelerationUncalibratedSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **dict(acc_uncal_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing acc_uncal_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and LinearAccelerationSeries in self._series) or self._series is None:
                try:
                    lin_acc_df = pd.read_sql_query("SELECT * from lin_acc_measurements_table", db_con)
                    self.measurements[LinearAccelerationSeries] = LinearAccelerationSeries(
                        filename=self.filename,
                        rdy_format_version=self.rdy_format_version,
                        **dict(lin_acc_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing lin_acc_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and MagnetometerSeries in self._series) or self._series is None:
                try:
                    mag_df = pd.read_sql_query("SELECT * from mag_measurements_table", db_con)
                    self.measurements[MagnetometerSeries] = MagnetometerSeries(filename=self.filename,
                                                                               rdy_format_version=self.rdy_format_version,
                                                                               **dict(mag_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing mag_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and MagnetometerUncalibratedSeries in self._series) or self._series is None:
                try:
                    mag_uncal_df = pd.read_sql_query("SELECT * from mag_uncal_measurements_table", db_con)
                    self.measurements[MagnetometerUncalibratedSeries] = MagnetometerUncalibratedSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **dict(mag_uncal_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing mag_uncal_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and OrientationSeries in self._series) or self._series is None:
                try:
                    orient_df = pd.read_sql_query("SELECT * from orient_measurements_table", db_con)
                    self.measurements[OrientationSeries] = OrientationSeries(filename=self.filename,
                                                                             rdy_format_version=self.rdy_format_version,
                                                                             **dict(orient_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing orient_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and GyroSeries in self._series) or self._series is None:
                try:
                    gyro_df = pd.read_sql_query("SELECT * from gyro_measurements_table", db_con)
                    self.measurements[GyroSeries] = GyroSeries(filename=self.filename,
                                                               rdy_format_version=self.rdy_format_version,
                                                               **dict(gyro_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing gyro_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and GyroUncalibratedSeries in self._series) or self._series is None:
                try:
                    gyro_uncal_df = pd.read_sql_query("SELECT * from gyro_uncal_measurements_table", db_con)
                    self.measurements[GyroUncalibratedSeries] = GyroUncalibratedSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **dict(gyro_uncal_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing gyro_uncal_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and RotationSeries in self._series) or self._series is None:
                try:
                    rot_df = pd.read_sql_query("SELECT * from rot_measurements_table", db_con)
                    self.measurements[RotationSeries] = RotationSeries(filename=self.filename,
                                                                       rdy_format_version=self.rdy_format_version,
                                                                       **dict(rot_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing rot_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and GPSSeries in self._series) or self._series is None:
                try:
                    gps_df = pd.read_sql_query("SELECT * from gps_measurements_table", db_con)
                    self.measurements[GPSSeries] = GPSSeries(filename=self.filename,
                                                             rdy_format_version=self.rdy_format_version,
                                                             **dict(gps_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing gps_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and GNSSMeasurementSeries in self._series) or self._series is None:
                try:
                    gnss_df = pd.read_sql_query("SELECT * from gnss_measurement_table", db_con)
                    self.measurements[GNSSMeasurementSeries] = GNSSMeasurementSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **dict(gnss_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing gnss_measurement_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and GNSSClockMeasurementSeries in self._series) or self._series is None:
                try:
                    gnss_clock_df = pd.read_sql_query("SELECT * from gnss_clock_measurement_table", db_con)
                    self.measurements[GNSSClockMeasurementSeries] = GNSSClockMeasurementSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **dict(gnss_clock_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing gnss_clock_measurement_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and NMEAMessageSeries in self._series) or self._series is None:
                try:
                    nmea_df = pd.read_sql_query("SELECT * from nmea_messages_table", db_con)
                    self.measurements[NMEAMessageSeries] = NMEAMessageSeries(
                        filename=self.filename, rdy_format_version=self.rdy_format_version, **dict(nmea_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing nmea_messages_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and PressureSeries in self._series) or self._series is None:
                try:
                    pressure_df = pd.read_sql_query("SELECT * from pressure_measurements_table", db_con)
                    self.measurements[PressureSeries] = PressureSeries(filename=self.filename,
                                                                       rdy_format_version=self.rdy_format_version,
                                                                       **dict(pressure_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing pressure_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and TemperatureSeries in self._series) or self._series is None:
                try:
                    temperature_df = pd.read_sql_query("SELECT * from temperature_measurements_table", db_con)
                    self.measurements[TemperatureSeries] = TemperatureSeries(filename=self.filename,
                                                                             rdy_format_version=self.rdy_format_version,
                                                                             **dict(temperature_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing temperature_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and HumiditySeries in self._series) or self._series is None:
                try:
                    humidity_df = pd.read_sql_query("SELECT * from humidity_measurements_table", db_con)
                    self.measurements[HumiditySeries] = HumiditySeries(filename=self.filename,
                                                                       rdy_format_version=self.rdy_format_version,
                                                                       **dict(humidity_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing humidity_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and LightSeries in self._series) or self._series is None:
                try:
                    light_df = pd.read_sql_query("SELECT * from light_measurements_table", db_con)
                    self.measurements[LightSeries] = LightSeries(filename=self.filename,
                                                                 rdy_format_version=self.rdy_format_version,
                                                                 **dict(light_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing light_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and WzSeries in self._series) or self._series is None:
                try:
                    wz_df = pd.read_sql_query("SELECT * from wz_measurements_table", db_con)
                    self.measurements[WzSeries] = WzSeries(filename=self.filename,
                                                           rdy_format_version=self.rdy_format_version,
                                                           **dict(wz_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug("(%s) DatabaseError occurred when accessing wz_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and SubjectiveComfortSeries in self._series) or self._series is None:
                try:
                    subjective_comfort_df = pd.read_sql_query("SELECT * from subjective_comfort_measurements_table",
                                                              db_con)
                    self.measurements[SubjectiveComfortSeries] = SubjectiveComfortSeries(
                        filename=self.filename,
                        rdy_format_version=self.rdy_format_version,
                        **dict(subjective_comfort_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing subjective_comfort_measurements_table" % self.filename)
                    logger.debug(e)

            if (self._series is not None and NTPDatetimeSeries in self._series) or self._series is None:
                try:
                    ntp_datetime_df = pd.read_sql_query("SELECT * from ntp_measurements_table", db_con)
                    self.measurements[NTPDatetimeSeries] = NTPDatetimeSeries(
                        filename=self.filename,
                        rdy_format_version=self.rdy_format_version,
                        strip_timezone=self.strip_timezone,
                        **dict(ntp_datetime_df))
                except (DatabaseError, PandasDatabaseError) as e:
                    logger.debug(
                        "(%s) DatabaseError occurred when accessing subjective_comfort_measurements_table" % self.filename)
                    logger.debug(e)

            db_con.close()

        else:
            raise ValueError("File extension %s is not supported" % self.extension)

        if self.cutoff:
            for m in self.measurements.values():
                if len(m) > 0:
                    m.cutoff(self.timestamp_when_started, self.timestamp_when_stopped)

    def to_df(self, interpolate: bool = True) -> pd.DataFrame:
        """ Merges the measurement series to a single DataFrame

        Parameters
        ----------
            interpolate : bool
                If true interpolates NaN values of concatenated measurement series

        Returns
        -------
            pd.DataFrame
        """
        data_frames = [series.to_df() for series in self.measurements.values()]

        # Concatenate dataframes and resort them ascending
        df_merged = pd.concat(data_frames).sort_index()

        # Merge identical indices by taking mean of column values and then interpolate NaN values
        if interpolate:
            df_merged = df_merged.groupby(level=0).mean().interpolate()
        else:
            df_merged = df_merged.groupby(level=0).mean()
        return df_merged


class FileIterator:
    def __init__(self, file: RDYFile):
        self._file = file
        self._series_types = list(self._file.measurements.keys())
        self._index = 0

    def __next__(self):
        if self._index < len(self._series_types):
            series_type = self._series_types[self._index]

            self._index += 1
            return self._file.measurements[series_type]
        else:
            raise StopIteration



from typing import Union

from pandas import Series


class Device:
    def __init__(self,
                 api_level: Union[int, Series] = -1,
                 base_os: Union[str, Series] = "N/A",
                 brand: Union[str, Series] = "N/A",
                 manufacturer: Union[str, Series] = "N/A",
                 device: Union[str, Series] = "N/A",
                 product: Union[str, Series] = "N/A",
                 model: Union[str, Series] = "N/A",
                 gnss_hardware_model_name: Union[str, Series] = "N/A",
                 gnss_year_of_hardware: Union[int, Series] = -1):
        """ Class representing Android device information

        Parameters
        ----------
        api_level
        base_os
        brand
        manufacturer
        device
        product
        model
        gnss_hardware_model_name
        gnss_year_of_hardware
        """
        if type(api_level) == Series:
            self.api_level = api_level[0] if len(api_level) > 0 else None
        else:
            self.api_level = api_level

        if type(base_os) == Series:
            self.base_os = base_os[0] if len(base_os) > 0 else None
        else:
            self.base_os = base_os

        if type(brand) == Series:
            self.brand = brand[0] if len(brand) > 0 else None
        else:
            self.brand = brand

        if type(manufacturer) == Series:
            self.manufacturer = manufacturer[0] if len(manufacturer) > 0 else None
        else:
            self.manufacturer = manufacturer

        if type(device) == Series:
            self.device = device[0] if len(device) > 0 else None
        else:
            self.device = device

        if type(product) == Series:
            self.product = product[0] if len(product) > 0 else None
        else:
            self.product = product

        if type(model) == Series:
            self.model = model[0] if len(model) > 0 else None
        else:
            self.model = model

        if type(gnss_hardware_model_name) == Series:
            self.gnss_hardware_model_name = gnss_hardware_model_name[0] if len(gnss_hardware_model_name) > 0 else None
        else:
            self.gnss_hardware_model_name = gnss_hardware_model_name

        if type(gnss_year_of_hardware) == Series:
            self.gnss_year_of_hardware = gnss_year_of_hardware[0] if len(gnss_year_of_hardware) > 0 else None
        else:
            self.gnss_year_of_hardware = gnss_year_of_hardware

    def __repr__(self):
        return "Brand: %s, Model: %s, Product: %s, Device: %s, Manufacturer: %s, Base OS: %s, API Level: %d, " \
               "GNSS Hardware Model Name: %s, GNSS Year of Hardware %d" \
               % (self.brand, self.model, self.product, self.device, self.manufacturer, self.base_os,
                  self.api_level if self.api_level else -1, self.gnss_hardware_model_name,
                  self.gnss_year_of_hardware if self.gnss_year_of_hardware else -1)



import math
from math import radians, cos, sin, asin, sqrt
from typing import Union, List, Tuple

import numpy as np
import overpy
import scipy.interpolate as si
from numpy.linalg import norm
from shapely.geometry import LineString

from pyridy import config


def calc_unit_vector(v: Union[list, np.ndarray]) -> np.ndarray:
    """ Returns the unit vector of the given vector v

    Parameters
    ----------
    v : array_like
        Vector of which the unit vector should be calculated
    """
    return v / np.linalg.norm(v)


def calc_angle_between(v1: Union[list, np.ndarray], v2: Union[list, np.ndarray]) -> np.ndarray:
    """ Returns the angle in radians between vectors v1 and v2

    Parameters
    ----------
    v1 : array_like
        First vector
    v2 : array_like
        Second vector
    """
    v1_u = calc_unit_vector(v1)
    v2_u = calc_unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def calc_curvature(x: List[float], y: List[float]) -> List[float]:
    """ Calculates the Menger curvature for a set of coordinates

    Parameters
    ----------
    x
    y

    Returns
    -------
    list
    """
    if len(x) != len(y):
        raise ValueError("x and y have to be same length")

    if len(x) > 0 and len(y) > 0:
        c = [0]

        for i in range(len(x)):
            if not (i == 0 or i == len(x) - 1):
                # Get three neighbored points
                x1 = x[i - 1]
                y1 = y[i - 1]

                x2 = x[i]
                y2 = y[i]

                x3 = x[i + 1]
                y3 = y[i + 1]

                # Get distance between each of the points
                s_a = math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
                s_b = math.sqrt((x2 - x3) ** 2 + (y2 - y3) ** 2)
                s_c = math.sqrt((x3 - x1) ** 2 + (y3 - y1) ** 2)

                s = (s_a + s_b + s_c) / 2
                a = s * (s - s_a) * (s - s_b) * (s - s_c)

                if a > 0:
                    A = math.sqrt(a)
                else:
                    A = 0

                # Calculate sign
                dx12 = x2 - x1
                dy12 = y2 - y1

                dx23 = x3 - x2
                dy23 = y3 - y2

                sgn = np.sign(np.cross([dx12, dy12], [dx23, dy23]))

                # Menger Curvature
                res = sgn * 4 * A / (s_a * s_b * s_c)
                if not math.isnan(res):
                    c.append(res)
                else:
                    c.append(0.0)

        c.append(0)
        return c
    else:
        return []


def calc_distance_from_xy(x: List[float], y: List[float]) -> Tuple[list, list]:
    """ Computes distances between a set of x y coordinates as well as the total distance

    Parameters
    ----------
    x
    y

    Returns
    -------
    Tuple
        List containing total distance, list of pairwise distances
    """
    if len(x) != len(y):
        raise ValueError("x and y have to be same length")

    if x and y:
        s = [0]
        ds = []

        for i in range(len(x)):
            if i >= 1:
                x1 = x[i - 1]
                y1 = y[i - 1]

                x2 = x[i]
                y2 = y[1]

                ds.append(math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2))

        # Integrate ds
        s.extend(np.cumsum(ds))

        return s, ds
    else:
        return [], []


def calc_distance_from_lon_lat(lon: List[float], lat: List[float]):
    """ Computes pairwise and total distance using geodesic distance of individual lat/lon coordinates

    Parameters
    ----------
    lon: list
    lat: list

    Returns
    -------
    list, list
    """
    if len(lon) != len(lat):
        raise ValueError("x and y have to be same length")

    if lon and lat:
        s = [0]
        ds = []

        for i in range(len(lon)):
            if i >= 1:
                ds.append(config.geod.inv(lon[i - 1], lat[i - 1], lon[i], lat[i])[2])

        # Integrate ds
        s.extend(np.cumsum(ds))

        return s, ds
    else:
        return [], []


def convert_way_to_line_string(w: overpy.Way, frmt: str = "lon,lat") -> LineString:
    if frmt == "lon,lat":
        coords = [(float(n.lon), float(n.lat)) for n in w.nodes]
        return LineString(coords)
    elif frmt == "x,y":
        lons, lats = [float(n.lon) for n in w.nodes], [float(n.lat) for n in w.nodes]
        xs, ys = config.proj(lons, lats)
        coords = [(x, y) for x, y in zip(xs, ys)]
        return LineString(coords)
    else:
        raise ValueError(f"Unknown format: {frmt}")


def convert_lon_lat_to_xy(lon: List[float], lat: List[float], adjust_zero_point: bool = False):
    """ Convert lon/lat coordinates to a metric coordinate system. The first coordinate is used for zero-point in
    the metric coordinate system

    Parameters
    ----------
    adjust_zero_point: bool
        If True, first coordinate will be 0,0
    lon: list[float]
    lat: list[float]

    Returns
    -------
    list, list
    """
    if lon and lat:
        x, y = config.proj(lon, lat)
        if adjust_zero_point:
            return [el - x[0] for el in x], [el - y[0] for el in y]
        else:
            return x, y
    else:
        return [], []


def bspline(cv, n=10000, degree=3, periodic=False):
    """ Calculate n samples on a bspline

        Parameters
        ----------
        cv: array_like
            Array ov control vertices
        n: int
            Number of samples to return
        degree: int
            Curve degree
        periodic: bool
            True - Curve is closed, False - Curve is open
        Returns
        -------
        np.ndarray

        """

    # If periodic, extend the point array by count+degree+1
    cv = np.asarray(cv)
    count = len(cv)

    if periodic:
        factor, fraction = divmod(count + degree + 1, count)
        cv = np.concatenate((cv,) * factor + (cv[:fraction],))
        count = len(cv)
        degree = np.clip(degree, 1, degree)

    # If opened, prevent degree from exceeding count-1
    else:
        degree = np.clip(degree, 1, count - 1)

    # Calculate knot vector
    kv = None
    if periodic:
        kv = np.arange(0 - degree, count + degree + degree - 1, dtype='int')
    else:
        kv = np.concatenate(([0] * degree, np.arange(count - degree + 1), [count - degree] * degree))

    # Calculate query range
    u = np.linspace(periodic, (count - degree), n)

    # Calculate result
    return np.array(si.splev(u, (kv, cv.T, degree))).T


def project_point_onto_line(line: Union[np.ndarray, list], point: Union[np.ndarray, list]) -> tuple:
    """

    Parameters
    ----------
    line: np.ndarray
        List of two points defining the line in the form of [[x1, y1],[x2, y2]]
    point: np.ndarray
        Point of which the distance perpendicular from the line should be calculated to
    Returns
    -------
    tuple
        Returns a tuple with the point where the orthogonal projections of the given points intersects the given the
        line and secondly the (perpendicular) distance to this point

    """
    if type(line) is list:
        line = np.array(line)

    if type(point) is list:
        point = np.array(point)

    p1 = line[0]
    p2 = line[1]
    p3 = point

    if np.array_equal(p1, p2):
        raise ValueError("Given line consists of two identical points!")

    d = norm(np.cross(p2 - p1, p1 - p3)) / norm(p2 - p1)

    n = p2 - p1
    n = n / norm(n, 2)

    p = p1 + n * np.dot(p3 - p1, n)

    return p, d


def boxes_to_edges(boxes):
    """
    Source: https://stackoverflow.com/questions/4842613/merge-lists-that-share-common-elements

        treat `l` as a Graph and returns it's edges
        to_edges(['a','b','c','d']) -> [(a,b), (b,c),(c,d)]
    """
    it = iter(boxes)
    last = next(it)

    for current in it:
        yield last, current
        last = current


def overlap(b1: List[float], b2: List[float], thres: float = .8) -> bool:
    """ Check whether to bounding boxes have an overlap larger than thres
        Each box must be of form x_min, y_min, x_max, y_max

    Parameters
    ----------
    thres
    b1
    b2
    """
    a1 = abs(b1[2] - b1[0]) * abs(b1[3] - b1[1])
    a2 = abs(b2[2] - b2[0]) * abs(b2[3] - b2[1])

    if b2[0] > b1[2] or b2[1] > b1[3]:  # Boxes not overlapping
        return False
    else:
        inter = (b1[2] - b2[0]) * (b1[3] - b2[1])

        r1 = inter / a1 if a1 > 0 else -1.0
        r2 = inter / a2 if a2 > 0 else -1.0

        if r1 > thres or r2 > thres:
            return True
        else:
            return False


def iou(b1: List[float], b2: List[float]):
    """ Calculates the Intersection over Union metric for two bounding boxes
        Each box must be of form x_min, y_min, x_max, y_max

    Parameters
    ----------
    b1
    b2
    """
    b1_x0, b1_y0 = b1[0], b1[1]
    b1_x1, b1_y1 = b1[2], b1[3]

    b2_x0, b2_y0 = b2[0], b2[1]
    b2_x1, b2_y1 = b2[2], b2[3]

    b1_w = b1[2] - b1[0]
    b1_h = b1[3] - b1[1]

    b2_w = b2[2] - b2[0]
    b2_h = b2[3] - b2[1]

    a1 = abs(b1_w) * abs(b1_h)
    a2 = abs(b2_w) * abs(b2_h)

    x0 = max(min(b1_x0, b1_x1), min(b2_x0, b2_x1))
    y0 = max(min(b1_y0, b1_y1), min(b2_y0, b2_y1))
    x1 = min(max(b1_x0, b1_x1), max(b2_x0, b2_x1))
    y1 = min(max(b1_y0, b1_y1), max(b2_y0, b2_y1))

    if x1 < x0 or y1 < y0:
        return 0.0

    inter = (x1 - x0) * (y1 - y0)
    union = a1 + a2 - inter

    iou = inter / union if union > 0 else -1.0

    assert 0.0 <= iou <= 1.0
    return iou


def is_point_within_line_projection(line: Union[np.ndarray, list], point: Union[np.ndarray, list]) -> bool:
    """ Checks whether a given points line projection falls within the points that the define the line

    Parameters
    ----------
    line: np.ndarray
        List of two points defining the line in the form of [[x1, y1],[x2, y2]]
    point: np.ndarray
        Point of which it should be determined whether the projection onto the line falls within the points that
        define the line
    Returns
    -------

    """
    if type(line) is list:
        line = np.array(line)

    if type(point) is list:
        point = np.array(point)

    p1 = line[0]
    p2 = line[1]
    p3 = point

    s = p2 - p1
    v = p3 - p1

    b = (0 <= np.inner(v, s) <= np.inner(s, s))

    return b


def haversine(lon1: float, lat1: float, lon2: float, lat2: float):
    """ Calculate the great circle distance in kilometers between two points on the earth (specified in decimal degrees)
        Source: https://stackoverflow.com/questions/4913349/haversine-formula-in-python-bearing-and-distance-between-two-gps-points
    Parameters
    ----------
    lon1 : float
        Longitude of first point
    lat1 : float
        Latitude of first point
    lon2 : float
        Longitude of second point
    lat2 : float
        Latitude of second point
    Returns
    -------
    float
        Distance between the points in meters

    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    r = 6371000  # Radius of earth in meters
    return c * r



class Sensor:
    def __init__(self,
                 name: str = None,
                 vendor: str = None,
                 string_type: str = None,
                 power: float = None,
                 resolution: float = None,
                 version: int = None,
                 type: int = None,
                 max_delay: int = None,
                 max_range: float = None,
                 min_delay: int = None):
        """ Class that describes an Android sensor

        See https://developer.android.com/reference/android/hardware/Sensor for documentation on individual parameters

        Parameters
        ----------
        name: str
        vendor: str
        string_type: str
        power: float
        resolution: float
        version: int
        type: int
        max_delay: int
        max_range: int
        min_delay: int
        """
        self.name = name
        self.vendor = vendor
        self.string_type = string_type
        self.power = power
        self.resolution = resolution
        self.version = version
        self.type = type
        self.max_delay = max_delay
        self.max_range = max_range
        self.min_delay = min_delay
        pass

    def __repr__(self):
        return self.string_type + " " + self.name



import itertools
import logging
from abc import ABC
from typing import List

import networkx as nx
import overpy
from ipyleaflet import Map, ScaleControl, FullScreenControl, Polyline, Circle, LayerGroup

from pyridy import config
from pyridy.osm.utils import convert_lon_lat_to_xy, calc_curvature, calc_distance_from_lon_lat
from pyridy.utils.tools import generate_random_color

logger = logging.getLogger(__name__)


class OSMResultNode:
    def __init__(self, lon: float, lat: float,
                 value=None, f=None, proc=None, dir: str = "", color: str = None):
        """ Class representing a Node calculated by PyRidy

        Parameters
        ----------
        color: str
            Node colors
        lat: float
            Latitude of node coordinate
        lon: float
            Longitude of node coordinate
        """

        self.lat = lat
        self.lon = lon
        self.value = value
        self.f = f
        self.proc = proc
        self.dir = dir
        self.color = '#CC071E' if not color else color

    def __repr__(self):
        return "Result node at (%s, %s)" % (self.lon, self.lat)


class OSMResultWay:
    def __init(self, way, res: float = .5):
        self.way: overpy.Way = way
        self.res = res
        pass


class OSMRailwayElement(ABC):
    def __init__(self, n: overpy.Node):
        """ Abstract Base Class for railway elements retrieved from OpenStreetMap

        Parameters
        ----------
        n: overpy.Node
            OpenStreetMap node queried by Overpy
        """
        self.n = n
        self.attributes = n.attributes
        self.tags = n.tags
        self.lat = float(n.lat)
        self.lon = float(n.lon)
        self.id = n.id

        if hasattr(n, "ways"):
            self.ways = n.ways
        else:
            self.ways = None

        self.__dict__.update(n.tags)  # Different elements contain different tags


class OSMLevelCrossing(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """ Class representing railway level crossings

        See https://wiki.openstreetmap.org/wiki/Tag:railway%3Dlevel_crossing for more information on available tags

        Parameters
        ----------
        n: overpy.Node
            OpenStreetMap Node retrieved using Overpy
        """
        super(OSMLevelCrossing, self).__init__(n)

    def __repr__(self):
        return "Level Crossing at (%s, %s)" % (self.lon, self.lat)


class OSMRailwayMilestone(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """ Class representing railway milestones (turnouts)

        Parameters
        ----------
        n: overpy.Node
            OpenStreetMap node retrieved using Overpy
        """
        super(OSMRailwayMilestone, self).__init__(n)

        pos = n.tags.get("railway:position", "-1").replace(",", ".").split("+")

        try:
            self.position = float(pos[0])
        except ValueError:
            logger.debug("Unusual milestone position format: %s" % n.tags.get("railway:position", ""))
            self.position = None
        self.addition = "" if len(pos) == 1 else pos[1]

    def __repr__(self):
        return "Milestone at (%s, %s): %.3f" % (self.lon, self.lat, self.position)


class OSMRailwaySignal(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """ Class representing railway signals

        See https://wiki.openstreetmap.org/wiki/Tag:railway%3Dsignal for more information on available tags

        Parameters
        ----------
        n: overpy.Node
            OpenStreetMap node retrieved using Overpy
        """
        super(OSMRailwaySignal, self).__init__(n)

    def __repr__(self):
        return "Signal at (%s, %s)" % (self.lon, self.lat)


class OSMRailwaySwitch(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """ Class representing railway switches (turnouts)

        Parameters
        ----------
        n: overpy.Node
            OpenStreetMap node retrieved using Overpy
        """
        super(OSMRailwaySwitch, self).__init__(n)
        self.allowed_transits = []  # List of triples in form of (neighbor 1, switch id, neighbor 2)

    def __repr__(self):
        return "Switch at (%s, %s)" % (self.lon, self.lat)


class OSMRelation:
    def __init__(self, relation: overpy.Relation, ways=None, color=None):
        """ Class Representing an OpenStreetMap relation. A relation can represent multiple tracks in some cases

        Parameters
        ----------
        relation: overpy.Relation
            Relation
        ways: List[overpy.Way]
            Ways part of the relation
        color:
            Color when used to draw the track e.g. using ipyleaflet
        """

        if ways is None:
            ways = []

        self.id = relation.id
        self.relation = relation
        self.name = relation.tags.get("name", "")
        self.ways = ways
        self.way_nodes = [way.nodes for way in self.ways]  # List of list of nodes
        self.nodes = list(itertools.chain.from_iterable(self.way_nodes)) if self.way_nodes else None  # list of nodes

        self.G = nx.MultiGraph()
        self.G.add_nodes_from([(n.id, n.__dict__) for n in self.nodes])

        # Add edges
        for w in self.ways:
            edges = [(n1.id, n2.id, config.geod.inv(float(n1.lon), float(n1.lat), float(n2.lon), float(n2.lat))[2])
                     for n1, n2 in zip(w.nodes, w.nodes[1:])]  # Edges have geodesic distances as edge weights
            self.G.add_weighted_edges_from(edges, weight="d", way_id=w.id)

        # Look up endpoints
        self.endpoints = []  # Node IDs of endpoints
        for n in self.G.nodes:
            if len(self.G.adj[n]) == 1:
                self.endpoints.append(n)

        # Search tracks within relation (double tracks have 2 physical tracks but 4 tracks are found through Graph
        # search since each track can be trafficked in both directions)
        self.tracks = []

        for s, t in itertools.combinations(self.endpoints, 2):
            try:
                sp_n = nx.shortest_path(self.G, source=s, target=t)  # List of node ids that make up shortest path
                nodes = [next(n for n in self.nodes if n.id == n_id) for n_id in sp_n]
                ways = list(set(list(itertools.chain.from_iterable([n.ways for n in nodes]))))
                self.tracks.append(OSMTrack(nodes, ways))
            except nx.NetworkXNoPath as e:
                logger.debug(e)

        logger.debug("Number of individual tracks: %d" % len(self.tracks))

        self.color = relation.tags.get("colour", generate_random_color("HEX")) if not color else color
        self.lon_sw = min([float(n.lon) for n in self.nodes])
        self.lon_ne = max([float(n.lon) for n in self.nodes])
        self.lat_sw = min([float(n.lat) for n in self.nodes])
        self.lat_ne = max([float(n.lat) for n in self.nodes])

    def to_ipyleaflef(self) -> List[list]:
        """

        Returns
        -------

        """
        if self.nodes:
            logger.warning("Relation has no nodes!")
            return [[n.lat, n.lon] for n in self.nodes]
        else:
            return [[]]

    def create_map(self, show_result_nodes: bool = False, use_file_color: bool = False) -> Map:
        center = ((self.lat_sw + self.lat_ne) / 2, (self.lon_sw + self.lon_ne) / 2)

        m = Map(center=center, zoom=12, scroll_wheel_zoom=True, basemap=config.OPEN_STREET_MAP_DE)
        m.add_control(ScaleControl(position='bottomleft'))
        m.add_control(FullScreenControl())

        # Add map
        m.add_layer(config.OPEN_RAILWAY_MAP)

        for track in self.tracks:
            coords = track.to_ipyleaflet()
            file_polyline = Polyline(locations=coords, color=self.color, fill=False, weight=4)
            m.add_layer(file_polyline)

        if show_result_nodes:
            nodes = list(itertools.chain.from_iterable([w.attributes.get("results", []) for w in self.ways]))
            circles = []
            for n in nodes:
                circle = Circle()
                circle.location = n.lon, n.lat
                circle.radius = 2
                circle.color = n.f.color if use_file_color else n.color
                circle.fill_color = n.f.color if use_file_color else n.color
                circle.weight = 3
                circle.fill_opacity = 0.1
                circles.append(circle)

            l_circles = LayerGroup(layers=circles)
            m.add_layer(l_circles)

        return m


class OSMRailwayLine(OSMRelation):
    def __init__(self, relation: overpy.Relation, ways: List[overpy.Way] = None, color=None):
        """ Class representing a railway line

        See https://wiki.openstreetmap.org/wiki/Tag:railway%3Drail

        Parameters
        ----------
        id: int
            ID of the railway line
        ways: list
            List of ways of the railway line
        tags: dict
            Tags associated with the railway line
        members: list
            list of nodes and ways associated with the railway line
        """
        super(OSMRailwayLine, self).__init__(relation=relation, ways=ways, color=color)

        self.tags = relation.tags

        self.members = relation.members
        self.milestones = [OSMRailwayMilestone(n) for n in self.nodes if n.tags.get("railway", "") == "milestone"]
        self.results = {}

    def __repr__(self):
        return self.__dict__.get("name", "")


class OSMTrack:
    def __init__(self, nodes: List[overpy.Node], ways: List[overpy.Way]):
        """ Represents a single railway track

        Parameters
        ----------
        nodes: List[overpy.node]
            Nodes that make up the track
        """
        self.lat = []
        self.lon = []

        self.x = []
        self.y = []

        self.ds = []
        self.s = []
        self.c = []

        self.nodes = nodes
        self.ways = ways

    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, nodes: List[overpy.Node]):
        self._nodes = nodes

        self.lon = [float(n.lon) for n in nodes]
        self.lat = [float(n.lat) for n in nodes]

        self.x, self.y = convert_lon_lat_to_xy(self.lon, self.lat)
        self.c = calc_curvature(self.x, self.y)
        self.s, self.ds = calc_distance_from_lon_lat(self.lon, self.lat)

    def flip_curvature(self):
        """
            Flips the calculated curvature upside down
        """
        self.c = [el * -1 for el in self.c]

    def to_ipyleaflet(self):
        """ Converts the coordinates to the format required by ipyleaflet for drawing
        Returns
        -------
            list
        """
        if self.lat and self.lon:
            return [[float(lat), float(lon)] for lat, lon in zip(self.lat, self.lon)]
        else:
            return [[]]

    def to_tuple_list(self, frmt: str = "lon,lat"):
        """ Converts the coordinates to a list of tuples
        Parameters
        ----------
            frmt : str, default: "lon,lat"
                Format, lon,lat or x,y

        Returns
        -------
            list
        """
        if frmt == "lon,lat":
            if self.lat and self.lon:
                return [[float(lat), float(lon)] for lat, lon in zip(self.lat, self.lon)]
            else:
                return [(None, None)]
        elif frmt == "x,y":
            if self.x and self.y:
                return [(x, y) for x, y in zip(self.x, self.y)]
            else:
                return [(None, None)]
        else:
            raise ValueError("frmt %s not supported" % frmt)

    def __repr__(self):
        return "Track from %d to %d, Length: %f" % (self.nodes[0].id,
                                                    self.nodes[-1].id,
                                                    self.s[-1])
